@@ -4,7 +4,7 @@ from pathlib import Path
 
 from typer import Option, Argument
 
-from basis.cli.services.graph import find_graph_file
+from basis.cli.services.lookup import IdLookup
 from basis.cli.services.output import console
 from basis.graph.builder import graph_manifest_from_yaml
 
@@ -19,8 +19,8 @@ def manifest(
     graph: Path = Argument(None, exists=True, help=_graph_help),
 ):
     """Print a graph manifest as JSON"""
-    graph_path = find_graph_file(graph)
-    m = graph_manifest_from_yaml(graph_path, allow_errors=True)
+    ids = IdLookup(explicit_graph_path=graph)
+    m = graph_manifest_from_yaml(ids.graph_file_path, allow_errors=True)
     j = json.dumps(m.dict(exclude_none=True))
     if pretty:
         console.print_json(j)
