@@ -1,8 +1,5 @@
 from __future__ import annotations
 
-from itertools import chain
-from typing import List
-
 from basis.cli.services.api import Endpoints, get_json
 from basis.cli.services.pagination import paginated
 
@@ -15,10 +12,6 @@ def get_organization_by_id(organization_uid: str) -> dict:
     return get_json(Endpoints.organization_by_id(organization_uid))
 
 
-def list_organizations() -> List[dict]:
-    return list(chain.from_iterable(paginated_organizations()))
-
-
 @paginated
 def paginated_organizations():
     return get_json(Endpoints.ORGANIZATIONS_LIST)
