@@ -1,10 +1,11 @@
 from __future__ import annotations
 
 import dataclasses
+import functools
 from dataclasses import dataclass
 from pathlib import Path
 from typing import List, Tuple, Dict, Set, Optional, Union
-from itertools import chain
+
 from basis.configuration.base import load_yaml
 from basis.configuration.graph import GraphDefinitionCfg, NodeCfg, PortMappingCfg
 from basis.configuration.path import GraphEdge, NodeId, PortId
@@ -18,6 +19,8 @@ from basis.graph.configured_node import (
     PortType,
     GraphError,
     InputDefinition,
+    ResolvedParameterValue,
+    ParameterDefinition,
 )
 from basis.node.sql.jinja import parse_interface_from_sql
 from basis.utils.ast_parser import read_interface_from_py_node_file
@@ -32,9 +35,8 @@ def graph_manifest_from_yaml(
     return manifest
 
 
-class NodeParseException(RuntimeError):
-    def __init__(self, *args, **kwargs):  # real signature unknown
-        super().__init__(*args, **kwargs)
+class NodeParseException(Exception):
+    pass
 
 
 # per-graph node port name mapping and exposed edges
@@ -44,9 +46,8 @@ class _Mapping:
     outputs: Dict[str, Tuple[ConfiguredNode, OutputDefinition]]
     # dict of node id -> port name mapping from yaml
     inputs: Dict[NodeId, Dict[str, str]]
-    # dicts of exposed port name -> list of edges connecting to that port
-    local_edges_by_exposed_input: Dict[str, List[GraphEdge]]
-    local_edges_by_exposed_output: Dict[str, List[GraphEdge]]
+    # dict of parameter name -> definitions with that name
+    parameters: Dict[str, List[ParameterDefinition]]
 
 
 # node parse result
@@ -56,9 +57,9 @@ class _Interface:
     node_type: NodeType
     # empty for non-graph nodes
     local_edges: List[GraphEdge]
-    node_name: str = None
-    node_id: NodeId = None
-    relative_node_path: Path = None
+    node_name: Optional[str] = None
+    node_id: Optional[NodeId] = None
+    relative_node_path: Optional[str] = None
 
 
 # graph parse result
@@ -80,6 +81,7 @@ class _GraphBuilder:
     def build(self) -> GraphManifest:
         parse = self._parse_graph_cfg(self.root_graph_location, None)
         self._resolve_edges()
+        self._resolve_parameter_values()
         return GraphManifest(
             graph_name=parse.name or self.root_dir.name,
             manifest_version=CURRENT_MANIFEST_SCHEMA_VERSION,
@@ -87,15 +89,19 @@ class _GraphBuilder:
             errors=self.errors,
         )
 
-    def _err(self, node_or_id: Union[ConfiguredNode, NodeId], message: str):
+    def _err(self, node_or_id: Union[ConfiguredNode, NodeId, None], message: str):
         id = (
             node_or_id.id
             if isinstance(node_or_id, ConfiguredNode)
+            else None
+            if node_or_id is None
             else NodeId(node_or_id)
         )
         self.errors.append(GraphError(node_id=id, message=message))
 
     def _read_graph_yml(self, path: Path) -> GraphDefinitionCfg:
+        if path.name != "graph.yml":
+            raise ValueError(f"Invalid graph file name: {path.name}")
         try:
             text = path.read_text()
         except Exception:
@@ -112,39 +118,35 @@ class _GraphBuilder:
     ) -> _ParsedGraph:
         config = self._read_graph_yml(node_yaml_path)
         node_dir = node_yaml_path.parent
-        exposed_inputs = (
-            config.exposes.inputs if config.exposes and config.exposes.inputs else []
-        )
-        exposed_outputs = (
-            config.exposes.outputs if config.exposes and config.exposes.outputs else []
-        )
-        mapping = _Mapping(
-            {}, {}, {i: [] for i in exposed_inputs}, {i: [] for i in exposed_outputs}
-        )
+        mapping = _Mapping({}, {}, {})
 
         # step 1: parse all the interfaces
-        nodes = [
-            self._parse_node_entry(node, node_dir, parent, mapping)
-            for node in (config.nodes or [])
-        ]
+        nodes: List[ConfiguredNode] = []
+        for node in config.nodes or []:
+            parsed = self._parse_node_entry(node, node_dir, parent, mapping)
+            nodes.append(parsed)
+            self.nodes_by_id[parsed.id] = parsed
 
         # step 2: set local edges
-        self._set_local_input_edges(parent, nodes, mapping)
-        self._set_local_exposed_output_edges(parent, mapping)
+        def get_exposed(attr: str) -> Optional[List[str]]:
+            if config.exposes is None:
+                return None
+            return getattr(config.exposes, attr, [])
 
-        self.nodes_by_id.update({node.id: node for node in nodes})
+        exposed_inputs = get_exposed("inputs")
+        exposed_outputs = get_exposed("outputs")
+        exposed_params = get_exposed("parameters")
+
+        local_edges = self._set_local_input_edges(
+            parent, nodes, mapping, exposed_inputs
+        )
+        local_edges += self._set_local_exposed_output_edges(
+            parent, mapping, exposed_outputs
+        )
 
         interface = self._build_graph_interface(
-            nodes, mapping, exposed_inputs, exposed_outputs
+            nodes, mapping, exposed_inputs, exposed_outputs, exposed_params
         )
-        local_edges = list(
-            chain(
-                chain.from_iterable(mapping.local_edges_by_exposed_input.values()),
-                chain.from_iterable(mapping.local_edges_by_exposed_output.values()),
-            )
-        )
-        self._check_for_unconnected_exposed_inputs(parent, mapping)
-
         return _ParsedGraph(config.name, interface, local_edges)
 
     # create a NodeInterface from a graph node's exposed ports
@@ -152,29 +154,46 @@ class _GraphBuilder:
         self,
         nodes: List[ConfiguredNode],
         mapping: _Mapping,
-        exposed_inputs: List[str],
-        exposed_outputs: List[str],
+        exposed_inputs: Optional[List[str]],
+        exposed_outputs: Optional[List[str]],
+        exposed_params: Optional[List[str]],
     ) -> NodeInterface:
         input_defs_by_name = {
             mapping.inputs[n.id][i.name]: i for n in nodes for i in n.interface.inputs
         }
-        # Copy the graph port definitions from the originating node, changing the name to the exposed name
+        # Copy the graph port definitions from the originating node, changing the name
+        # to the exposed name
         outputs = [
             mapping.outputs[o][1].copy(update={"name": o})
-            for o in exposed_outputs
+            for o in exposed_outputs or mapping.outputs.keys()
             if o in mapping.outputs  # may be missing in broken graphs
         ]
         inputs = [
             input_defs_by_name[i].copy(update={"name": i})
             for i in exposed_inputs
+            or [v for d in mapping.inputs.values() for v in d.values()]
             if i in input_defs_by_name  # may be missing in broken graphs
         ]
-        return NodeInterface(inputs=inputs, outputs=outputs, parameters=[], state=None)
+        parameters = [
+            mapping.parameters[p][0]
+            for p in exposed_params or mapping.parameters.keys()
+            if p in mapping.parameters
+            and mapping.parameters[p]  # may be missing in broken graphs
+        ]
+        return NodeInterface(
+            inputs=inputs, outputs=outputs, parameters=parameters, state=None
+        )
 
     # set local edges connected to nodes' inputs
     def _set_local_input_edges(
-        self, graph_id: Optional[NodeId], nodes: List[ConfiguredNode], mapping: _Mapping
-    ):
+        self,
+        graph_id: Optional[NodeId],
+        nodes: List[ConfiguredNode],
+        mapping: _Mapping,
+        exposed_inputs: Optional[List[str]],  # None when all inputs are exposed
+    ) -> List[GraphEdge]:
+        edges = []
+        connected_exposed_inputs = set()
         for node in nodes:
             for i in node.interface.inputs:
                 name = mapping.inputs[node.id][i.name]
@@ -188,33 +207,44 @@ class _GraphBuilder:
                         node.local_edges,
                         src.local_edges,
                     )
-                elif name in mapping.local_edges_by_exposed_input:
+                elif (exposed_inputs is None and graph_id is not None) or (
+                    exposed_inputs and name in exposed_inputs
+                ):
+                    connected_exposed_inputs.add(name)
                     self._set_edge(
-                        graph_id,
-                        name,
-                        node.id,
-                        i.name,
-                        node.local_edges,
-                        mapping.local_edges_by_exposed_input[name],
+                        graph_id, name, node.id, i.name, node.local_edges, edges,
                     )
                 elif i.required:
                     self._err(node, f'Cannot find output named "{name}"')
                     self.broken_ports.add(PortId(node_id=node.id, port=i.name))
 
+        for name in exposed_inputs or []:
+            if name not in connected_exposed_inputs:
+                self._err(graph_id, f'Exposed input does not exist: "{name}"')
+        return edges
+
     # set local edges on any nodes connected to exposed outputs
     def _set_local_exposed_output_edges(
-        self, graph_id: Optional[NodeId], mapping: _Mapping
-    ):
-        for (name, edges) in mapping.local_edges_by_exposed_output.items():
+        self,
+        graph_id: Optional[NodeId],
+        mapping: _Mapping,
+        exposed_outputs: Optional[List[str]],  # None when all outputs are exposed
+    ) -> List[GraphEdge]:
+        if graph_id is None:
+            if exposed_outputs:
+                raise ValueError("Cannot expose outputs on root graph")
+            else:
+                return []
+        if exposed_outputs is None:
+            exposed_outputs = mapping.outputs
+        edges = []
+        for name in exposed_outputs:
             if name in mapping.outputs:
-                assert graph_id is not None
                 (src, o) = mapping.outputs[name]
                 self._set_edge(src.id, o.name, graph_id, name, edges, src.local_edges)
             elif graph_id is not None:
                 self._err(graph_id, f'Exposed output does not exist: "{name}"')
-            else:
-                # we'll have to revisit this if we want to support manifests for components
-                raise ValueError("Cannot expose outputs on root graph")
+        return edges
 
     def _set_edge(
         self,
@@ -236,18 +266,15 @@ class _GraphBuilder:
     def _check_edge(
         self, dst_id: NodeId, name: str, src_t: PortType, dst_t: PortType,
     ):
-        if src_t == PortType.Table and dst_t == PortType.Stream:
-            self._err(
-                dst_id,
-                f"Cannot connect {name}: input is a {src_t}, but output is a {dst_t}",
-            )
-
-    def _check_for_unconnected_exposed_inputs(
-        self, graph_id: Optional[NodeId], mapping: _Mapping
-    ):
-        for name, edges in mapping.local_edges_by_exposed_input.items():
-            if not edges:
-                self._err(graph_id, f'Exposed input does not exist: "{name}"')
+        if src_t == dst_t:
+            return
+        node_name = self.nodes_by_id[dst_id].name
+        self._err(
+            dst_id,
+            f"Cannot connect {name} to {node_name}: "
+            f"{node_name} expects an Input{dst_t.title()}, "
+            f"but {name} is an Output{src_t.title()}",
+        )
 
     # record all the output name mappings in this node's config
     def _track_outputs(
@@ -271,6 +298,12 @@ class _GraphBuilder:
             else:
                 mapping.outputs[name] = (node, o)
 
+    def _track_parameters(
+        self, parameters: List[ParameterDefinition], mapping: _Mapping
+    ):
+        for p in parameters:
+            mapping.parameters[p.name] = mapping.parameters.get(p.name, []) + [p]
+
     def _track_inputs(
         self,
         node_id: NodeId,
@@ -283,13 +316,23 @@ class _GraphBuilder:
         mapping.inputs[node_id] = names
 
     def _resolve_edges(self):
+        inputs_by_node_and_output = {}
+        for n in self.nodes_by_id.values():
+            if n.node_type == NodeType.Graph:
+                for e in n.local_edges:
+                    inputs_by_node_and_output[(n.id, e.output)] = e.input
         for node in self.nodes_by_id.values():
             if node.node_type == NodeType.Graph:
                 continue
             for local_edge in node.local_input_edges():
-                self._resolve_edge(local_edge, node)
+                self._resolve_edge(local_edge, node, inputs_by_node_and_output)
 
-    def _resolve_edge(self, local_edge: GraphEdge, node: ConfiguredNode):
+    def _resolve_edge(
+        self,
+        local_edge: GraphEdge,
+        node: ConfiguredNode,
+        inputs_by_node_and_output: Dict[Tuple[NodeId, PortId], PortId],
+    ):
         vertex_port = local_edge.input
         vertex = self.nodes_by_id[vertex_port.node_id]
 
@@ -297,9 +340,7 @@ class _GraphBuilder:
             return  # already reported an error for this edge
 
         while vertex.node_type == NodeType.Graph:
-            vertex_port = next(
-                (e.input for e in vertex.local_edges if e.output == vertex_port), None,
-            )
+            vertex_port = inputs_by_node_and_output.get((vertex.id, vertex_port))
             if not vertex_port:
                 self._err(
                     node, f"Could not resolve edge for port {local_edge.input.port}"
@@ -327,6 +368,38 @@ class _GraphBuilder:
             local_edge.output.node_id, local_edge.output.port, src_t, dst_t
         )
 
+    def _resolve_parameter_values(self):
+        @functools.lru_cache(maxsize=None)
+        def get_param_values(node_id: NodeId) -> Dict[str, ResolvedParameterValue]:
+            node = self.nodes_by_id[node_id]
+            exposed = {p.name for p in node.interface.parameters}
+            d = {
+                k: ResolvedParameterValue(value=v, source=node_id)
+                for (k, v) in node.parameter_values.items()
+            }
+
+            # check for setting unexposed values
+            for k in node.parameter_values.keys():
+                if k not in exposed:
+                    self._err(
+                        node_id,
+                        f'No exposed parameter named "{k}" on node "{node.name}"',
+                    )
+
+            # walk up the tree, overriding values with any set in parents
+            parent_id = node.parent_node_id
+            if parent_id:
+                parent_values = get_param_values(parent_id)
+                parameters = {*d.keys(), *(p.name for p in node.interface.parameters)}
+                for k in parameters:
+                    if k in parent_values and k in exposed:
+                        d[k] = parent_values[k]
+            node.resolved_parameter_values.update(d)
+            return {k: v for k, v in d.items() if k in exposed}
+
+        for n in self.nodes_by_id.values():
+            get_param_values(n.id)
+
     def _default_node_name(self, path: Path) -> str:
         # For 'graph.yml' files, use the directory name
         if path.suffix in (".yml", ".yaml"):
@@ -354,7 +427,17 @@ class _GraphBuilder:
             raise ValueError(f"Must specify 'node_file' or 'webhook' for all nodes")
 
         if interface.node_id in self.nodes_by_id:
-            raise ValueError(f"Duplicate node id: {interface.node_id}")
+            existing = self.nodes_by_id[interface.node_id]
+            existing_path = existing.file_path_to_node_script_relative_to_root
+            if existing.name == interface.node_name:
+                s = f"name '{existing.name}'"
+            else:
+                s = f"id '{existing.id}'"
+
+            raise ValueError(
+                f"Duplicate node {s} for nodes {existing_path} "
+                f"and {interface.relative_node_path}"
+            )
 
         configured_node = ConfiguredNode(
             name=interface.node_name,
@@ -368,6 +451,7 @@ class _GraphBuilder:
             schedule=node.schedule,
             local_edges=interface.local_edges,
             resolved_edges=[],  # we'll fill these in once we have all the nodes parsed
+            resolved_parameter_values={},
         )
 
         self._track_outputs(
@@ -376,6 +460,7 @@ class _GraphBuilder:
         self._track_inputs(
             interface.node_id, interface.node.inputs, node.inputs, mapping
         )
+        self._track_parameters(interface.node.parameters, mapping)
         return configured_node
 
     def _parse_webhook_node_entry(
